/*
 *  connld.c - unique pipe generator
 *
 *  On the first open of this module, a connection to it's stream head
 *  is made, but nothing else.  On subsequent calls, however, the open
 *  does the following:
 *  . generate a unique pipe
 *  . return one endpoint of the pipe to replace itself, i.e, by returning
 *    the near endpoint's device number
 *  . send the far endpoint's file via an M_PASSFP message to the
 *    the far endpoint of its own pipe, to be picked up via an
 *    I_RECVFD ioctl.
 *  Thus, this module can only be passed on a pipe (not even on
 *  a single-ended FIFO).
 *
 *  Note that this module will never handle message traffic.
 *
 *  Copyright (C)  2000  John A. Boyd Jr.  protologos, LLC
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 * 
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the
 *  Free Software Foundation, Inc., 59 Temple Place - Suite 330, Cambridge,
 *  MA 02139, USA.
 */

#ident "@(#) LiS connld.c 1.4 2/6/01"

#ifdef MODULE
#  if defined(LINUX) && defined(__KERNEL__)
#    ifdef MODVERSIONS
#      include <linux/modversions.h>
#    endif
#    include <linux/module.h>
#  else
#    error This can only be a module in the Linux kernel environment
#  endif
#else
#  define MOD_INC_USE_COUNT  do {} while (0)
#  define MOD_DEC_USE_COUNT  do {} while (0)
#endif

#include <sys/LiS/config.h>

#include <sys/stream.h>
#include <sys/stropts.h>
#include <sys/errno.h>
#include <sys/cmn_err.h>

#include <sys/LiS/head.h>
#include <sys/osif.h>

/*
 *  Some configuration sanity checks
 */
#ifndef CONNLD_
#error Not configured
#endif

#ifndef CONNLD__ID
#define CONNLD__ID 0xaabc
#endif

#define  MOD_ID   CONNLD__ID
#define  MOD_NAME "connld"

/*
 *  function prototypes
 */
static int  connld_open(queue_t *, dev_t*, int, int, cred_t *);
static int  connld_close(queue_t *, int, cred_t *);

/*
 *  module structure
 */
static struct module_info connld_minfo =
{ 
    MOD_ID,            /* id */
    MOD_NAME,          /* name */
    0,                 /* min packet size accepted */
    INFPSZ,            /* max packet size accepted */
    10240L,            /* highwater mark */
    512L               /* lowwater mark */
};

static struct qinit connld_rinit = {
    NULL,               /* put */
    NULL,               /* service */
    connld_open,        /* open */
    connld_close,       /* close */
    NULL,               /* admin */
    &connld_minfo,      /* info */
    NULL                /* stat */
};

static struct qinit connld_winit = { 
    NULL,               /* put */
    NULL,               /* service */
    NULL,               /* open */
    NULL,               /* close */
    NULL,               /* admin */
    &connld_minfo,      /* info */
    NULL                /* stat */
};

struct streamtab connld_info =
{
    &connld_rinit,      /* read queue init */
    &connld_winit,      /* write queue init */
    NULL,               /* lower mux read queue init */
    NULL                /* lower mux write queue init */
};

/*
 *  open
 */
static int connld_open( q, devp, flag, sflag, credp )
queue_t *q;
dev_t *devp;
int flag, sflag;
cred_t *credp;
{
    stdata_t *head = q->q_str;

    if (!head || head->magic != STDATA_MAGIC) {
	cmn_err( CE_CONT,
		 "connld_open(): bad reference to stream head\n" );
	return -EINVAL;
    }
    if (!(head->sd_flag & STFIFO)) {
	cmn_err( CE_CONT,
		 "connld_open(): not even a FIFO!!\n" );
	return -EPIPE;
    }
    if (!head->sd_peer || head->sd_peer == head) {
	cmn_err( CE_CONT,
		 "connld_open(): not a valid pipe\n" );
	return -EPIPE;
    }

#ifdef DEBUG
    cmn_err(
	CE_CONT,
	"connld_open( 0x%x, 0x%x, 0x%x,0x%x, ... ) #%d head 0x%x file 0x%x\n",
	q, *devp, flag, sflag, (int)q->q_ptr, (int)head, (int)head->sd_file );
#endif

    if (q->q_ptr) {
	struct file *f0, *f1, *oldf = head->sd_file;
	int error;

	if (head != FILE_STR(oldf)) {
	    cmn_err( CE_CONT,
		     "connld_open(): invalid sd_file 0x%x\n", head->sd_file );
	    return -EINVAL;
	}
#ifdef DEBUG
	cmn_err( CE_CONT,
		 "connld_open(): opening file 0x%x\n", head->sd_file );
#endif
	
	if ((error = lis_get_pipe( &f0, &f1 )) < 0) {
	    cmn_err( CE_CONT,
		     "connld_open(): get_pipe() failed: %d\n", error );
	    return error;
	}

#ifdef DEBUG
	cmn_err( CE_CONT,
		 "connld_open(): created pipe 0x%x/0x%x <-> 0x%x/0x%x\n",
		 FILE_INODE(f0)->i_rdev, FILE_STR(f0),
		 FILE_INODE(f1)->i_rdev, FILE_STR(f1) );
#endif
	
	/*
	 *  send one of the new pipe's ends to the peer of the connld pipe
	 */
	if ((error = lis_sendfd( head, -1, f1 )) < 0) {
	    cmn_err( CE_CONT,
		     "connld_open(): sendfd( 0x%x,-1,0x%x ) failed - %d\n",
		     error );
	    lis_strclose( FILE_INODE(f0), f0 );
	    lis_strclose( FILE_INODE(f1), f1 );
	    return error;
	}
	/*
	 *  now replace the current opening file's inode, and
	 *  discard the extra file pointer
	 */

	if (!lis_old_inode( head->sd_file, FILE_INODE(f0) )) {
	    lis_strclose( FILE_INODE(f0), f0 );
	    return -EINVAL;
	}

	/*
	 *  the current stream has changed - we've attached the other pipe
	 *  end to it instead, so the pipe end now has two references.
	 *  The old stream was disposed of in the swap; now bump the pipe
	 *  end's ref count, and close it once.
	 */	
	*devp = FILE_INODE(f0)->i_rdev;
	head->sd_refcnt--;

	head = FILE_STR(f0);
	SET_FILE_STR(oldf, head);
	head->sd_refcnt++;

	lis_strclose( FILE_INODE(f0), f0 );  /* close one reference */


#ifdef DEBUG
	cmn_err(
	    CE_CONT,
	    "connld_open( 0x%x, 0x%x, 0x%x,0x%x, ... ) #%d head 0x%x OK\n",
	    q, *devp, flag, sflag, (int)q->q_ptr, (int)head );
#endif

	return 0;
    } else {
	/*
	 *  first time in - check STCONNLD flag, then set it
	 */
	if (F_ISSET(head->sd_flag,STCONNLD)) {
	    cmn_err( CE_CONT,
		     "connld_open(): previously pushed 'connld'\n" );
	    return -EINVAL;
	}
	F_SET(head->sd_flag,STCONNLD);

	/*
	 *  set q_ptr as an open count flag
	 */
	((int)q->q_ptr)++;

	MOD_INC_USE_COUNT;
    
	return 0;
    }
}

/*
 *  close
 */
static int connld_close( q, flag, credp )
queue_t *q;
int flag;
cred_t *credp;
{
    stdata_t *head = q->q_str;

#ifdef DEBUG
    cmn_err( CE_CONT,
	     "connld_close( 0x%x, 0x%x, ... ) head 0x%x\n",
	     q, flag, credp, head );
#endif

    q->q_ptr = NULL;
#ifdef MODULE
    MOD_DEC_USE_COUNT;
#endif
    qprocsoff(q);
	
    if (!head || head->magic != STDATA_MAGIC) {
	cmn_err( CE_CONT,
		 "connld_open(): bad reference to stream head\n" );
	return -EINVAL;
    }
    F_CLR(head->sd_flag,STCONNLD);

    return 0;
}

#ifdef MODULE

/*
 *  Linux loadable module interface
 */

int init_module(void)
{
    int ret = lis_register_strmod( &connld_info, MOD_NAME );
    if (ret < 0) {
	cmn_err( CE_CONT,
		 "%s - unable to register module.\n",
		 MOD_NAME );
	return ret;
    }

    return 0;
}

void cleanup_module(void)
{
    if (lis_unregister_strmod(&connld_info) < 0)
	cmn_err( CE_CONT,
		 "%s - unable to unregister moduler.\n",
		 MOD_NAME );
    return;
}

#endif
