/************************************************************************
*                        LiS Memory Interface                           *
*************************************************************************
*									*
* This interface was created to insulate the Linux STREAMS programmer	*
* from the constantly changing memory allocation interface provided by	*
* the Linux kernel.							*
*									*
*    Copyright (C) 2000  David Grothe, Gcom, Inc <dave@gcom.com>	*
*									*
* This library is free software; you can redistribute it and/or		*
* modify it under the terms of the GNU Library General Public		*
* License as published by the Free Software Foundation; either		*
* version 2 of the License, or (at your option) any later version.	*
* 									*
* This library is distributed in the hope that it will be useful,	*
* but WITHOUT ANY WARRANTY; without even the implied warranty of	*
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU	*
* Library General Public License for more details.			*
* 									*
* You should have received a copy of the GNU Library General Public	*
* License along with this library; if not, write to the			*
* Free Software Foundation, Inc., 59 Temple Place - Suite 330,		*
* Cambridge, MA 02139, USA.						*
*									*
************************************************************************/

#ident "@(#) LiS lismem.c 1.5 11/22/00"

#include <sys/stream.h>		/* gets all the right LiS stuff included */
#include <sys/lismem.h>		/* LiS mem header file */
#include <sys/osif.h>		/* LiS kernel interface */
#include <sys/lislocks.h>	/* for spin locks */

/************************************************************************
*                          Page Allocation                              *
*************************************************************************
*									*
* Page allocation is the way to ensure that you get memory that		*
* occupies physically contiguous locations.  We offer a simplified	*
* interface to the kernel's page allocator here.  The kernel's page	*
* allocation routines are awkward to use in that you have to pass	*
* the number of pages to de-allocate to their "free pages" routine.	*
*									*
* We keep a separate table of allocated pages.  This allows us to return*
* the entire page to the caller.					*
*									*
************************************************************************/

typedef struct contig_memlink
{				/* 32 bytes long */
    struct contig_memlink *link;/* to next entry */
    void	*page_addr ;	/* addr of allocated page */
    int		 size ;		/* original size in bytes */
    int		 order ;	/* "order" param for freeing */
    char	*file ;		/* who allocated */
    int		 line ;
    int		 boundary ;	/* page boundary */

} contig_memlink_t ;

static contig_memlink_t	*contig_mem_head ;
static unsigned long	contig_bytes ;		/* total allocated */
static lis_spin_lock_t	contig_lock ;		/* protect the table */
static int		contig_init ;

#define chk_init						\
	if (!contig_init)					\
	{							\
	    lis_spin_lock_init(&contig_lock, "LiS-Page-Alloc");	\
	    contig_init = 1 ;					\
	}

static int	alloc_page_tbl(void)
{
    contig_memlink_t	*pg ;
    contig_memlink_t	*p ;
    int			 nentries ;

    pg = (contig_memlink_t *) get_free_page(GFP_KERNEL) ;
    if (pg == NULL) 
	return(-1) ;

    memset(pg, 0, PAGE_SIZE) ;
    for (p = pg, nentries = PAGE_SIZE/sizeof(*p); nentries > 0; nentries--, p++)
    {
	p->link = p + 1 ;
    }

    /*
     * Link the last entry in this page to the head entry of the previous page.
     * Then make this page the new head of the list.
     */
    pg->boundary = 1 ;			/* 1st entry is on page boundary */
    (p-1)->link = contig_mem_head ;	/* last entry in page */
    contig_mem_head = pg ;
    return(0) ;
}

static contig_memlink_t *find_entry(void *addr)
{
    contig_memlink_t	*mp ;

    for (mp = contig_mem_head; mp != NULL; mp = mp->link)
    {
	if (mp->page_addr == addr) return(mp) ;
    }

    return(NULL) ;			/* could not find it */
}

void    *lis_get_free_pages_fcn(int nbytes, char *file, int line)
{
    void		*a ;
    int			 order ;
    int			 bytes ;
    contig_memlink_t	*mp ;

    chk_init ;

    lis_spin_lock(&contig_lock) ;
    mp = find_entry(NULL) ;		/* find available entry */
    if (mp == NULL)
    {
	if (alloc_page_tbl() < 0)
	{
	    lis_spin_unlock(&contig_lock) ;
	    return(NULL) ;		/* can't track allocated page */
	}

	mp = find_entry(NULL) ;		/* should be one now */
	if (mp == NULL)			/* busted code */
	{
	    lis_spin_unlock(&contig_lock) ;
	    printk("lis_get_free_pages: called from %s #%d: "
		    "bug in allocation table code\n",
		    file, line) ;
	    return(NULL) ;
	}
    }

    for (order = 0; (bytes = (1 << order) * PAGE_SIZE) < nbytes; order++) ;

    a =  (void *) __get_free_pages(GFP_KERNEL, order) ;	/* kernel routine */
    if (a == NULL)
    {
	printk("lis_get_free_pages: called from %s #%d: "
		"nbytes=%d order=%d total=%lu allocation failed\n",
		file, line,
		nbytes, order, contig_bytes) ;
	lis_spin_unlock(&contig_lock) ;
	return(NULL) ;
    }

    contig_bytes += bytes ;		/* # bytes allocated */

    mp->page_addr    = a ;
    mp->size         = nbytes ;
    mp->order        = order ;
    mp->file         = file ;
    mp->line         = line ;

    lis_spin_unlock(&contig_lock) ;
    return(a) ;

} /* lis_get_free_pages_fcn */

void     *lis_free_pages_fcn(void *ptr, char *file, int line)
{
    contig_memlink_t	*mp ;
    void		*addr ;
    int			 order ;

    chk_init ;

    lis_spin_lock(&contig_lock) ;
    mp = find_entry(ptr) ;
    if (mp == NULL)			/* have no record of this page */
    {
	printk("lis_free_pages: called from %s #%d: "
	       "cannot find page info for address 0x%lx\n",
	       file, line, (long) ptr) ;
	lis_spin_unlock(&contig_lock) ;
	return(NULL) ;
    }

    addr = mp->page_addr ;
    order = mp->order ;
    contig_bytes -= (1 << mp->order) * PAGE_SIZE ;
    mp->page_addr = NULL ;

    lis_spin_unlock(&contig_lock) ;

    free_pages((unsigned long) addr, order) ; /* kernel routine */

    return(NULL) ;

} /* lis_free_pages_fcn */

void	lis_free_all_pages(void)
{
    contig_memlink_t	*mp ;
    contig_memlink_t	*boundary_ptr = NULL ;

    chk_init ;

    if (contig_bytes != 0)
    {
	lis_spin_lock(&contig_lock) ;
	/*
	 * Free the data pages allocated to callers.
	 */
	for (mp = contig_mem_head; mp != NULL; mp = mp->link)
	{
	    printk("lis_free_all_pages: "
	    	      "Memory at %lx allocated from %s #%d, "
		      "%d bytes leaked (recovered)\n",
		      (long) mp->page_addr, mp->file, mp->line, mp->size) ;
	    lis_free_pages(mp->page_addr) ;
	}

	/*
	 * Free the pages allocated for keeping track of other pages.
	 */
	for (mp = contig_mem_head; mp != NULL; mp = mp->link)
	{
	    if (!mp->boundary) continue ;

	    /* At a page boundary, remember it for freeing */
	    if (boundary_ptr != NULL)
		free_page((unsigned long) boundary_ptr) ;

	    boundary_ptr = mp ;		/* remember for next time */
	}

	if (boundary_ptr != NULL)	/* catch last page */
	    free_page((unsigned long) boundary_ptr) ;

	contig_mem_head = NULL ;

	lis_spin_unlock(&contig_lock) ;
    }

} /* lis_free_all_pages */


/************************************************************************
*                       Calls to Kernel Routines                        *
*************************************************************************
*									*
* We are really using the LiS memory allocator here so that we can	*
* track file and line number information.				*
*									*
************************************************************************/

void	*lis_alloc_atomic_fcn(int nbytes, char *file, int line)
{
    return(lis_malloc(nbytes, GFP_ATOMIC, file, line)) ;
}

void	*lis_alloc_kernel_fcn(int nbytes, char *file, int line)
{
    return(lis_malloc(nbytes, GFP_KERNEL, file, line)) ;
}

void	*lis_alloc_dma_fcn(int nbytes, char *file, int line)
{
    return(lis_malloc(nbytes, GFP_DMA, file, line)) ;
}

void	*lis_free_mem_fcn(void *mem_area, char *file, int line)
{
    lis_free(mem_area, file, line) ;
    return(NULL) ;
}

