/*                               -*- Mode: C -*- 
 * wait.c --- wait queues management
 * Author          : Francisco J. Ballesteros
 * Created On      : Tue May 31 22:25:19 1994
 * Last Modified By: David Grothe
 * RCS Id          : $Id: wait.c,v 1.2 1995/12/27 01:02:53 dave Exp $
 * Purpose         : keep close all the wait related stuff.
 * ----------------______________________________________________
 *
 *   Copyright (C) 1995  Francisco J. Ballesteros
 *   Copyright (C) 1997  David Grothe, Gcom, Inc <dave@gcom.com>
 *
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330, Cambridge,
 * MA 02139, USA.
 * 
 *
 *    You can reach me by email to
 *    nemo@ordago.uc3m.es
 */

#ident "@(#) LiS wait.c 2.16 3/13/01 22:02:12 "


/*  -------------------------------------------------------------------  */
/*				 Dependencies                            */

#include <sys/strport.h>
#include <sys/strconfig.h>	/* config definitions */
#include <sys/LiS/share.h>	/* streams shared defs*/
#include <sys/LiS/wait.h>	/* interface */
#include <sys/LiS/head.h>	/* stream head */
#include <sys/LiS/queue.h>	/* streams queues */
#include <sys/osif.h>


/*  -------------------------------------------------------------------  */
/*				   Symbols                               */

#define	AINC(v)		lis_atomic_inc(&(v))
#define	ADEC(v)		lis_atomic_dec(&(v))

/*  -------------------------------------------------------------------  */
/*				 Global vars                             */
long	lis_wakeup_close_wt_cnt ;
long	lis_sleep_on_close_wt_cnt ;

extern lis_atomic_t	 lis_in_syscall ;	/* in head.c */

/*  -------------------------------------------------------------------  */
/*			Exported functions & macros                      */

/* The functions below are to achieve mutual exclusion on the stream
 * head. They use the sd_wopen wait queue to sleep and the sd_lock char to
 * flag the lock.
 * The lock/unlock are the complete ones.
 * if you just want to wait you may use wait_on/wake_up.
 *
 * You can lock(L)/wait(W)/sleep(S) on the following events:
 * L/W wopen:  the 1st opener completes its job
 * W/S wwrite: there's room in the write queue to issue a downstream msg.
 * STATUS: complete(?), untested
 *
 * When we are about to sleep, and we have the stream head unlocked,
 * check to see if we need to run the queues.  Often we sleep after having
 * sent something downstream to the driver, so this is a good place
 * to check for queues to run.
 */
/*  -------------------------------------------------------------------  */
int 
lis_sleep_on_close_wt(void *q_str)
{
    stdata_t	*sd = (stdata_t *) q_str ;
    int		 rslt ;

    if (sd == NULL || sd->magic != STDATA_MAGIC)
	return(-ENOENT) ;

    lis_spin_unlock(&sd->sd_lock) ;		/* unlock strm head */
    lis_clear_and_save_sigs(sd);		/* undo pending signals */
    lis_sleep_on_close_wt_cnt++ ;
    lis_atomic_dec(&lis_in_syscall) ;		/* "done" with a system call */
    lis_runqueues();
    rslt = lis_down(&sd->sd_close_wt);
    lis_atomic_inc(&lis_in_syscall) ;		/* processing a system call */
    lis_restore_sigs(sd);			/* restore old signals */
    lis_spin_lock(&sd->sd_lock) ;		/* relock strm head */
    return(rslt) ;

}/*lis_sleep_on_close_wt*/
/*  -------------------------------------------------------------------  */
void 
lis_wakeup_close_wt(void *q_str)
{
    stdata_t	*sd = (stdata_t *) q_str ;

    if (   sd != NULL
	&& sd->magic == STDATA_MAGIC
	&& F_ISSET(sd->sd_flag,(STRCLOSEWT|STRFLUSHWT))
       )
    {
	lis_wakeup_close_wt_cnt++ ;
	lis_up(&sd->sd_close_wt);
    }

}/*lis_wakeup_close_wt*/
/*  -------------------------------------------------------------------  */
int 
lis_sleep_on_wioc(struct stdata * sd)
{
    int		rslt ;

    lis_spin_unlock(&sd->sd_lock) ;		/* unlock strm head */
    lis_atomic_dec(&lis_in_syscall) ;		/* "done" with a system call */
    lis_runqueues();
    rslt = lis_down(&sd->sd_wioc);
    lis_atomic_inc(&lis_in_syscall) ;		/* processing a system call */
    lis_spin_lock(&sd->sd_lock) ;		/* relock strm head */
    return(rslt) ;

}/*lis_sleep_on_wioc*/
/*  -------------------------------------------------------------------  */
int 
lis_sleep_on_wwrite(struct stdata * sd)
{
    int		ret ;

    lis_spin_unlock(&sd->sd_lock) ;		/* unlock strm head */
    AINC(sd->sd_wrcnt) ;
    lis_atomic_dec(&lis_in_syscall) ;		/* "done" with a system call */
    lis_runqueues();
    if ((ret = lis_down(&sd->sd_wwrite)) < 0)
	ADEC(sd->sd_wrcnt) ;

    lis_atomic_inc(&lis_in_syscall) ;		/* processing a system call */
    lis_spin_lock(&sd->sd_lock) ;		/* relock strm head */
    return(ret) ;

}/*lis_sleep_on_wwrite*/
/*  -------------------------------------------------------------------  */
void 
lis_wake_up_wwrite(struct stdata * sd)
{
    if (lis_atomic_read(&sd->sd_wrcnt) > 0)
    {
	ADEC(sd->sd_wrcnt) ;
	lis_up(&sd->sd_wwrite);
    }

}/*lis_wake_up_wwrite*/
/*  -------------------------------------------------------------------  */
void 
lis_wake_up_all_wwrite(struct stdata * sd)
{
    while (lis_atomic_read(&sd->sd_wrcnt) > 0)
    {
	ADEC(sd->sd_wrcnt) ;
	lis_up(&sd->sd_wwrite);
    }

}/*lis_wake_up_all_wwrite*/
/*  -------------------------------------------------------------------  */
int 
lis_sleep_on_wread(struct stdata * sd)
{
    int		ret ;

    AINC(sd->sd_rdcnt) ;			/* inc while q locked */
    lis_unlockq(sd->sd_rq) ;			/* unlock to let rsrv run */
    lis_spin_unlock(&sd->sd_lock) ;		/* unlock strm head */
    lis_atomic_dec(&lis_in_syscall) ;		/* "done" with a system call */
    lis_runqueues();
    if ((ret = lis_down(&sd->sd_wread)) < 0)	/* now sleep */
	ADEC(sd->sd_rdcnt) ;

    lis_atomic_inc(&lis_in_syscall) ;		/* processing a system call */
    lis_spin_lock(&sd->sd_lock) ;		/* relock strm head */
    lis_lockq(sd->sd_rq) ;			/* relock b4 returning */
    return(ret) ;

}/*lis_sleep_on_wread*/
/*  -------------------------------------------------------------------  */
void 
lis_wake_up_wread(struct stdata * sd)
{
    if (lis_atomic_read(&sd->sd_rdcnt) > 0)		/* someone is reading */
    {
	ADEC(sd->sd_rdcnt) ;
	lis_up(&sd->sd_wread);
    }

}/*lis_wake_up_wread*/
/*  -------------------------------------------------------------------  */
void 
lis_wake_up_all_wread(struct stdata * sd)
{
    while (lis_atomic_read(&sd->sd_rdcnt) > 0)
    {
	ADEC(sd->sd_rdcnt) ;
	lis_up(&sd->sd_wread);
    }

}/*lis_wake_up_all_wread*/

/*  -------------------------------------------------------------------  */
int 
lis_sleep_on_wiocing(struct stdata * sd)
{
    int		ret ;

    lis_spin_unlock(&sd->sd_lock) ;		/* unlock strm head */
    lis_atomic_dec(&lis_in_syscall) ;		/* "done" with a system call */
    lis_runqueues();
    ret = lis_down(&sd->sd_wiocing);
    lis_atomic_inc(&lis_in_syscall) ;		/* processing a system call */
    lis_spin_lock(&sd->sd_lock) ;		/* relock strm head */
    return(ret) ;

}/*lis_sleep_on_wiocing*/




/*----------------------------------------------------------------------
# Local Variables:      ***
# change-log-default-name: "~/src/prj/streams/src/NOTES" ***
# End: ***
  ----------------------------------------------------------------------*/
